#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   Philips 
@Time        :   2023/9/14 15:41
@Author      :   Xuesong Chen
@Description :   
"""
from wuji.Reader.EDF.Base import Base


class PhilipsEDFReader(Base):
    def __init__(self, file_path):
        super().__init__(file_path)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   NSRR 
@Time        :   2023/8/17 16:47
@Author      :   Xuesong Chen
@Description :   
"""
from wuji.Reader.EDF.Base import Base
import neurokit2 as nk


class NSRREDFReader(Base):
    def __init__(self, file_path):
        super().__init__(file_path)


if __name__ == '__main__':
    fp = '/Users/cxs/project/OSAPillow/data/SHHS/edfs/shhs1-200001.edf'
    reader = NSRREDFReader(fp)
    print(reader.get_signal(type='ecg', tmax=10))
    # stages = Reader.get_standard_sleep_stages()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

.. autosummary::
    :toctree: generated/

    mcnemar

"""

import numpy


def mcnemar(reference, estimated_a, estimated_b):
    """McNemar's test

    Wikipedia entry https://en.wikipedia.org/wiki/McNemar%27s_test

    Parameters
    ----------
    reference : list
        Reference value

    estimated_a : list
        System output A

    estimated_b : list
        System output B

    Returns
    -------
    float
        exact P-value

    """

    if len(reference) != len(estimated_a) or len(reference) != len(estimated_b):
        raise ValueError('Input arrays needs to be same length.')

    # Convert lists into numpy.array
    reference = numpy.array(reference)
    estimated_a = numpy.array(estimated_a)
    estimated_b = numpy.array(estimated_b)

    # Intermediate values
    correct_a = estimated_a == reference
    correct_b = estimated_b == reference

    incorrect_a = estimated_a != reference
    incorrect_b = estimated_b != reference

    # Contingency table values
    b = float( numpy.sum( numpy.logical_and(incorrect_a, correct_b) ) )
    c = float( numpy.sum( numpy.logical_and(correct_a, incorrect_b) ) )

    # Continuity corrected version of the McNemar test to approximate the binomial exact-P-value
    # Edwards, A (1948). "Note on the "correction for continuity" in testing the significance of the difference
    # between correlated proportions". Psychometrika. 13: 185–187
    if b + c > 0:
        return (numpy.abs(b - c) - 1)**2 / (b + c)
    else:
        return 0



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   HSPV2
@Time        :   2023/9/14 15:41
@Author      :   Xuesong Chen
@Description :
"""
import re

import matplotlib.pyplot as plt
import numpy as np

from wuji.Reader.EDF.Base import Base
import mne
from pyedflib.edfreader import EdfReader


class HSPV2EDFReader(Base):
    def __init__(self, file_path):
        super().__init__(file_path)

    def _parse_file(self, file_path):
        n_channels = len(mne.io.read_raw_edf(file_path, verbose='error').ch_names)
        self.reader = mne.io.read_raw_edf(file_path, misc=range(n_channels), verbose='error')
        self.signal_labels = np.array(self.reader.ch_names)
        self.duration = self.reader.n_times / self.reader.info['sfreq']
        self._assign_signal_types()

    def _assign_signal_types(self):
        self.signal_type = []
        for idx, sig in enumerate(self.signal_labels):
            if re.search('E[CK]G', sig, re.IGNORECASE):
                self.signal_type.append('ecg')
            elif re.search('S[pa]O2', sig, re.IGNORECASE):
                self.signal_type.append('spo2')
            elif re.search('ABD', sig, re.IGNORECASE):
                self.signal_type.append('abd')
            elif re.search('CHEST|THO', sig, re.IGNORECASE):
                self.signal_type.append('chest')
            elif re.search('C3-M2|C4-M1|F3-M2|F4-M1|O1-M2|O2-M1|C3|C4|F3|F4|O1|O2', sig, re.IGNORECASE):
                self.signal_type.append('eeg')
            elif re.search('EMG', sig, re.IGNORECASE):
                self.signal_type.append('emg')
            elif re.search('EOG', sig, re.IGNORECASE):
                self.signal_type.append('eog')
            elif re.search('Snore', sig, re.IGNORECASE):
                self.signal_type.append('snore')
            elif re.search('position', sig, re.IGNORECASE):
                self.signal_type.append('position')
            elif re.search('Flow|NEW AIR|AirFlow', sig, re.IGNORECASE):
                self.signal_type.append('flow')
            elif re.search('Numeric Aux', sig, re.IGNORECASE):
                self.signal_type.append('trigger')
            elif re.search('Pressure', sig, re.IGNORECASE):
                self.signal_type.append('nasal_pressure')
            elif re.search('Pleth', sig, re.IGNORECASE):
                self.signal_type.append('ppg')
            else:
                self.signal_type.append('unk')
        self.signal_type = np.array(self.signal_type)

    def get_start_recording_time(self):
        return self.reader.info['meas_date']

    def get_sample_frequency(self, **kwargs):
        return int(self.reader.info['sfreq'])

    def get_signal(self, ch_name=None, type='ecg', tmin=None, tmax=None, order=0):
        '''
        获取所有匹配信号，order代表次序，0代表第一个匹配的信号
        :param ch_name:
        :param type:
        :param tmin:
        :param tmax:
        :param order:
        :return:
        '''
        if ch_name:
            idx = np.argwhere(self.signal_labels == ch_name).flatten()[order]
        else:
            idx = np.argwhere(self.signal_type == type).flatten()[order]
        return self.reader.get_data(picks=idx, tmin=tmin, tmax=tmax).squeeze() * 1e6


if __name__ == '__main__':
    fp = '/Users/cxs/Downloads/sub-S0001111239981_ses-3_task-psg_eeg.edf'
    reader = HSPV2EDFReader(fp)
    import yasa
    eeg_chan_name = reader.get_channel_name(type='eeg')
    for tar_pattern in ['C', 'F']:
        for _ch_name in eeg_chan_name:
            if re.search(tar_pattern, _ch_name, re.IGNORECASE):
                print(_ch_name)
    yasa.SleepStaging(reader.reader, )
    # edf_reader = Base(fp)
    # base_ecg = edf_reader.get_signal(type='eeg', order=2)
    print(reader.get_channel_name(type='eeg'))
    plt.plot(reader.get_signal(type='eeg', order=2), label='mne_ecg')
    # plt.plot(base_ecg+10, label='base_ecg')
    plt.legend()
    plt.show()
    print()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   cut_edf 
@Time        :   2023/11/7 16:23
@Author      :   Xuesong Chen
@Description :   
"""

import pyedflib

def cut_edf(source_path, target_path, start_time, end_time):
    """
    Cut a segment of an EDF file and save it to a new file.
    Parameters
    ----------
    source_path
    target_path
    start_time: 单位为s
    end_time: 单位为s
    Returns
    -------

    """

    # 读取 EDF 文件
    with pyedflib.EdfReader(source_path) as edf_reader:
        n_channels = edf_reader.signals_in_file
        signal_headers = edf_reader.getSignalHeaders()
        n_samples = edf_reader.getNSamples()

        # 根据源文件类型创建新 EDF 文件
        file_type = pyedflib.FILETYPE_EDFPLUS if edf_reader.datarecord_duration <= 1 else pyedflib.FILETYPE_BDFPLUS
        with pyedflib.EdfWriter(target_path, n_channels, file_type=file_type) as edf_writer:
            edf_writer.setSignalHeaders(signal_headers)

            # Prepare a list to hold data for all channels
            data_list = []

            for i in range(n_channels):
                start_index = int(start_time * edf_reader.getSampleFrequency(i))
                end_index = int(end_time * edf_reader.getSampleFrequency(i))

                if end_index > n_samples[i]:  # Adjust if the end index is beyond the sample count
                    end_index = n_samples[i]

                # Read the data segment for each channel
                data = edf_reader.readSignal(i, start=start_index, n=end_index - start_index)
                data_list.append(data)

            # Write all channels at once
            edf_writer.writeSamples(data_list)

    print(f"new edf was stored in {target_path}. ")


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   NightlySleepMetrics 
@Time        :   2023/9/26 14:06
@Author      :   Xuesong Chen
@Description :   
"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import pearsonr
from sklearn.metrics import r2_score as r2_score_sklearn, recall_score, f1_score


class NightlySleepMetricsEvaluator:
    def __init__(self, gt, pred):
        self.gt = np.asarray(gt)
        self.pred = np.asarray(pred)

    def pearsonr(self, plot=False, ax=None):
        r, p_val = pearsonr(self.gt, self.pred)
        if plot is False:
            return round(r, 3)
        else:
            return_ax = True
            if ax is None:
                fig, ax = plt.subplots()
                return_ax = False
            ax.plot(self.gt, self.pred, 'o', color='blue', alpha=0.6)
            ax.set_xlabel('ground truth')
            ax.set_ylabel('prediction')
            ax.set_title(f'Pearson correlation coefficient: {round(r, 3)}')
            min_v = min(min(self.gt), min(self.pred))
            max_v = max(max(self.gt), max(self.pred))
            ax.plot([min_v, max_v], [min_v, max_v], '--', color='darkgray')
            if return_ax is False:
                plt.show()
            else:
                return ax

    def r2_score(self):
        r_squared = r2_score_sklearn(self.gt, self.pred)
        return round(r_squared, 3)

    def bland_altman_plot(self, ax=None):
        mean = np.mean([self.gt, self.pred], axis=0)
        diff = self.pred - self.gt
        md = np.mean(diff)
        sd = np.std(diff, axis=0)
        return_ax = True
        if ax is None:
            fig, ax = plt.subplots()
            return_ax = False
        # 使用渐变色来绘制散点图

        # 绘制散点图
        ax.scatter(mean, diff)

        # 绘制平均线和1.96倍标准差线
        ax.axhline(md, color='salmon', linestyle='-', label="Mean")
        ax.axhline(md + 1.96 * sd, color='lightsalmon', linestyle='--', label="Mean + 1.96 SD")
        ax.axhline(md - 1.96 * sd, color='lightsalmon', linestyle='--', label="Mean - 1.96 SD")

        ax.set_xlabel('Mean Score')
        ax.set_ylabel('Diff Score')
        ax.set_title('Bland-Altman Plot')
        ax.legend(loc='best')

        if not return_ax:
            plt.show()
        else:
            return ax

    def screen_OSA(self, threshold=15):
        gt_indices = np.where(self.gt > threshold, True, False)
        pred_indices = np.where(self.pred > threshold, True, False)
        sensitivity = recall_score(gt_indices, pred_indices)
        specificity = recall_score(np.logical_not(gt_indices), np.logical_not(pred_indices))
        f_score = f1_score(gt_indices, pred_indices)
        return {
            'sensitivity': round(sensitivity, 3),
            'specificity': round(specificity, 3),
            'f1_score': round(f_score, 3)
        }




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   AnnotationBase 
@Time        :   2023/8/17 16:21
@Author      :   Xuesong Chen
@Description :   
"""
from abc import ABC, abstractmethod
from wuji.Reader.utils import get_irequal_duration_chunks
import pandas as pd
import statistics
import numpy as np

dataset = 'common'


def merge_drop_duplicates_sort(*args, sort_column='Start'):
    result_df = pd.concat(args)
    if result_df.empty:
        return result_df
    result_df = result_df.drop_duplicates()
    result_df = result_df.sort_values(by=sort_column)
    result_df = result_df.reset_index(drop=True)
    return result_df


def add_not_scored_into_sleep_stages(df):
    if df.empty:
        return df
    rows = []

    # 遍历原始 DataFrame
    for i, row in df.iterrows():
        # 如果不是第一行且当前行的 Start 与上一行的 End 不相等
        if i > 0 and row['Start'] > df.iloc[i - 1]['End']:
            # 创建新行
            new_row = {
                'Type': 'NotScored',
                'Start': df.iloc[i - 1]['End'],
                'Duration': row['Start'] - df.iloc[i - 1]['End'],
                'End': row['Start']
            }
            rows.append(new_row)
        # 添加当前行
        rows.append(row)

    # 使用 pd.concat 创建新的 DataFrame
    new_df = pd.concat([pd.DataFrame([r]) for r in rows], ignore_index=True)
    return new_df


def assign_sleep_stage_label_to_AH_events(AH_events_raw, sleep_stages_raw):
    AH_events = AH_events_raw.copy()
    sleep_stages = sleep_stages_raw.copy()
    sleep_stages['End'] = sleep_stages['Start'] + sleep_stages['Duration']
    if dataset == 'hsp':
        sleep_stages = add_not_scored_into_sleep_stages(sleep_stages)
    legal_stages = {'N1', 'N2', 'N3', 'REM', 'Wake'}
    n_legal_nan = 0
    if not AH_events.empty:
        AH_events['End'] = AH_events['Start'] + AH_events['Duration']
        for idx, row in sleep_stages.iterrows():
            start_time = row['Start']
            end_time = row['End']
            AH_events.loc[
                (AH_events['Start'] >= start_time) & (AH_events['End'] <= end_time),
                'SleepStage'] = row['Type']
        sleep_stage_idx_cache = 0
        duration = sleep_stages.iloc[-1]['End']
        start_sleep_time = sleep_stages.iloc[0]['Start']
        AH_events['SleepStage'] = AH_events['SleepStage'].replace('nan', np.nan)
        for resp_idx, resp_row in AH_events[AH_events['SleepStage'].isna()].iterrows():
            if resp_row['Start'] > duration or resp_row['End'] < start_sleep_time:
                n_legal_nan += 1
                continue
            for sleep_stage_idx, sleep_stage_row in sleep_stages.loc[sleep_stage_idx_cache:].iterrows():
                if resp_row['End'] > sleep_stage_row['End'] and resp_row['Start'] < sleep_stage_row['End']:
                    overlap_on_cur_stage = sleep_stage_row['End'] - resp_row['Start']
                    overlap_on_next_stage = resp_row['End'] - sleep_stage_row['End']
                    stage_set = {sleep_stage_row['Type'],
                                 sleep_stages.loc[min(sleep_stage_idx + 1, len(sleep_stages) - 1), 'Type']}
                    stage_set = stage_set.intersection(legal_stages)
                    stage_set = list(stage_set)
                    # 如果当前事件对应的stage缺失，则取附近10个事件的众数
                    if len(stage_set) == 0:
                        cur_event_stage = statistics.mode(
                            sleep_stages.loc[
                            max(sleep_stage_idx - 5, 0):min(sleep_stage_idx + 5, len(sleep_stages)), 'Type'].values)
                        if cur_event_stage not in legal_stages:
                            cur_event_stage = np.nan
                            n_legal_nan += 1
                    # 如果当前事件对应的stage只有一个，且为Wake，则取nan
                    elif len(stage_set) == 1 and 'Wake' in stage_set:
                        cur_event_stage = np.nan
                        n_legal_nan += 1
                    # 如果当前事件对应的stage只有一个，则直接取该stage
                    elif len(stage_set) == 1 and 'Wake' not in stage_set:
                        cur_event_stage = stage_set[0]
                    # 如果当前事件对应的stage有两个，且其中一个为Wake，则取另一个stage
                    elif len(stage_set) != 1 and 'Wake' in stage_set:
                        stage_set.remove('Wake')
                        cur_event_stage = stage_set[0]
                    # 如果当前事件对应的stage有两个，且两个都不是Wake，则取重叠时间较长的stage
                    elif overlap_on_cur_stage >= overlap_on_next_stage:
                        cur_event_stage = sleep_stage_row['Type']
                    else:
                        cur_event_stage = sleep_stages.loc[sleep_stage_idx + 1, 'Type']
                    AH_events.loc[resp_idx, 'SleepStage'] = cur_event_stage
                    sleep_stage_idx_cache = sleep_stage_idx
                    break

        assert AH_events[
                   'SleepStage'].isna().sum() == n_legal_nan, f"存在未标记的呼吸事件{AH_events[AH_events['SleepStage'].isna()]}"
        AH_events = AH_events[~AH_events['SleepStage'].isna()]
    return AH_events


class AHEventFilter:
    def __init__(self, respiratory_events, od_events, arousal_events=None, sleep_stages=None):
        self.respiratory_events = respiratory_events
        self.od_events = od_events
        self.sleep_stages = sleep_stages
        self.arousal_events = arousal_events

    def _assign_sleep_stage_label_to_AH_events(self):
        self.AH_events = assign_sleep_stage_label_to_AH_events(self.AH_events, self.sleep_stages)

    def get_AH_with_OD(self, row, od_eps=45):
        if row['Type'] == 'Hypopnea':
            hypopnea_start = row['Start']
            hypopnea_end = hypopnea_start + row['Duration']

            # Check if this hypopnea event has already been processed
            if 'Processed_od' not in self.od_events.columns:
                self.od_events['Processed_od'] = False

            condition = ((hypopnea_start <= self.od_events['Start']) &
                         (hypopnea_end + od_eps > self.od_events['Start']) &
                         (~self.od_events['Processed_od']))
            rows_to_drop = self.od_events[condition].index

            if any(condition):
                # rows_to_drop可能有多个, [0] 将第一个匹配到的 OD 不再匹配后续其他的 H
                self.od_events.loc[rows_to_drop[0], 'Processed_od'] = True
                return True  # Break out of the loop if a match is found
            return False

        else:
            return True

    def get_AH_with_Arousal(self, row, aro_eps=6):
        if row['Type'] == 'Hypopnea':
            hypopnea_start = row['Start']
            hypopnea_end = hypopnea_start + row['Duration']

            if 'Processed_arousal' not in self.arousal_events.columns:
                self.arousal_events['Processed_arousal'] = False

            condition = ((hypopnea_start <= self.arousal_events['Start']) &
                         (hypopnea_end + aro_eps > self.arousal_events['Start']) &
                         (hypopnea_end < (self.arousal_events['Start'] + self.arousal_events['Duration'])) &
                         (~self.arousal_events['Processed_arousal']))

            rows_to_drop = self.arousal_events[condition].index

            if any(condition):
                # 第一个匹配到的 arousal 不再匹配后续其他的 H
                self.arousal_events.loc[rows_to_drop[0], 'Processed_arousal'] = True
                return True
            return False

        else:
            return True

    def get_filtered_AH_events(self, type='AHI3', od_eps=45, aro_eps=6):

        if type == 'AHI4':
            ah_with_od = self.respiratory_events[
                self.respiratory_events.apply(self.get_AH_with_OD, od_eps=od_eps, axis=1)]
            self.AH_events = ah_with_od
        elif type == 'AHI3':
            ah_with_od = self.respiratory_events.loc[
                self.respiratory_events.apply(self.get_AH_with_OD, od_eps=od_eps, axis=1)]
            ah_with_arousal = self.respiratory_events.loc[
                self.respiratory_events.apply(self.get_AH_with_Arousal, aro_eps=aro_eps, axis=1)]
            self.AH_events = merge_drop_duplicates_sort(ah_with_od, ah_with_arousal)
        self._assign_sleep_stage_label_to_AH_events()
        if self.AH_events.empty:
            return self.AH_events
        return self.AH_events[self.AH_events['SleepStage'].isin(['N1', 'N2', 'N3', 'REM'])]


class AHI:
    def __init__(self, AH_events, sleep_stages):
        if AH_events.empty:
            self.AH_events = AH_events
            return
        assert 'SleepStage' in AH_events.columns, 'AH_events中缺少SleepStage列'
        self.total_sleep_time = sleep_stages[~sleep_stages['Type'].isin(['Wake', 'NotScored'])]['Duration'].sum()
        self.total_sleep_time_in_hours = self.total_sleep_time / 3600
        self.total_sleep_time_in_hours_REM = sleep_stages[sleep_stages['Type'].isin(['REM'])]['Duration'].sum() / 3600
        self.total_sleep_time_in_hours_NREM = sleep_stages[sleep_stages['Type'].isin(['N1', 'N2', 'N3'])][
                                                  'Duration'].sum() / 3600
        self.sleep_stages = get_irequal_duration_chunks(sleep_stages)
        self.AH_events = AH_events.copy()
        self.sleep_stages['End'] = self.sleep_stages['Start'] + self.sleep_stages['Duration']
        self.AH_events['End'] = self.AH_events['Start'] + self.AH_events['Duration']

    def get_AHI(self, type='Total'):

        if self.AH_events.empty:
            return 0

        if type == 'Total':
            ahi = self.AH_events['SleepStage'].isin(
                ['N1', 'N2', 'N3', 'REM']).sum() / self.total_sleep_time_in_hours
        elif type == 'REM':
            ahi = self.AH_events['SleepStage'].isin(['REM']).sum() / self.total_sleep_time_in_hours_REM
        elif type == 'NREM':
            ahi = self.AH_events['SleepStage'].isin(
                ['N1', 'N2', 'N3']).sum() / self.total_sleep_time_in_hours_NREM
        else:
            raise ValueError("Invalid type. Must be 'Total', 'REM', or 'NREM'.")

        return ahi


class Base(ABC):
    def __init__(self, file_path):
        self.sleep_stages = None
        self.respiratory_events = None
        self.AH_events = None
        self.OD_events = None
        self._parse_file(file_path)

    @abstractmethod
    def _parse_file(self, file_path):
        self.recording_start_time = None
        self.duration = None
        self.anno_df = None

    def get_recording_start_time(self):
        return self.recording_start_time

    def get_duration(self):
        return self.duration

    def get_sleep_onset_time(self):
        if self.sleep_stages is None:
            self.sleep_stages = self.get_standard_sleep_stages()
        onset_time = self.sleep_stages[self.sleep_stages['Type'].isin(['N1', 'N2', 'N3', 'REM'])]['Start'].values[0]
        return onset_time

    def total_sleep_time(self):
        if self.sleep_stages is None:
            self.sleep_stages = self.get_standard_sleep_stages()
        total_sleep_time = self.sleep_stages[self.sleep_stages['Type'].isin(['N1', 'N2', 'N3', 'REM'])][
            'Duration'].sum()
        return total_sleep_time

    def get_standard_sleep_stages(self):
        '''
        用于获取标准的睡眠分期标记

        | Type | Start | Duration |
        |------|-------|----------|

        Type: Wake, N1, N2, N3, REM
        Start: 从睡眠开始到当前分期的时间
        Duration: 当前分期的持续时间，统一为30s

        :return:
        上述Dataframe格式
        '''
        return None

    def get_standard_AH_events(self):
        '''
        用于获取标准的呼吸暂停、低通气标记

        | Type | Start | Duration |
        |------|-------|----------|

        Type: Apnea, Hypopnea
        Start: 从睡眠开始到当前事件的开始时间
        Duration: 当前事件的持续时间

        :return:
        上述Dataframe格式
        '''
        pass

    def plot_sleep_stage(self, ax=None):
        import matplotlib.pyplot as plt
        from matplotlib.dates import DateFormatter

        # 设置开始时间
        start_time = self.recording_start_time
        sleep_stage = self.get_standard_sleep_stages() if self.sleep_stages is None else self.sleep_stages

        # 按照指定的顺序更新类型映射
        type_order = ['N3', 'N2', 'N1', 'REM', 'Wake', 'NotScored']
        type_mapping = {type: i for i, type in enumerate(type_order)}
        # 创建颜色映射，根据每个类型的实际含义选择颜色
        color_mapping = {
            "N3": "darkblue",
            "N2": "purple",
            "N1": "blue",
            "REM": "green",
            "Wake": "lightblue",
            "NotScored": "grey"
        }

        sleep_stage['Start'] = pd.to_timedelta(sleep_stage['Start'], unit='s') + start_time
        sleep_stage['End'] = sleep_stage['Start'] + pd.to_timedelta(sleep_stage['Duration'], unit='s')

        # 创建图表
        if not ax:
            fig, ax = plt.subplots(figsize=(100, 5))

        prev_end = None
        prev_type = None
        for _, row in sleep_stage.iterrows():
            ax.barh(type_mapping[row['Type']], row['End'] - row['Start'], left=row['Start'], height=0.5,
                    align='center',
                    color=color_mapping[row['Type']])
            if prev_end is not None and prev_type is not None:
                ax.plot([prev_end, row['Start']], [type_mapping[prev_type], type_mapping[row['Type']]],
                        color='lightgrey')
            prev_end = row['End']
            prev_type = row['Type']

        # 设置y轴的标签
        ax.set_yticks(range(len(type_mapping)))
        ax.set_yticklabels(list(type_mapping.keys()))

        # 设置x轴的刻度标签为实际的小时和分钟
        ax.xaxis.set_major_formatter(DateFormatter('%H:%M'))
        # set x limits
        start_time = sleep_stage['Start'].values[0]
        end_time = sleep_stage['End'].values[-1]
        ax.set_xlim([start_time, end_time])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   xxi2hr 
@Time        :   2024/8/27 17:24
@Author      :   Xuesong Chen
@Description :   
"""

import numpy as np
from scipy.interpolate import interp1d
from numpy.typing import NDArray
from hrvanalysis import get_nn_intervals
import pandas as pd


def interpolate_hr(hr: NDArray, duration: int, target_fs: int) -> NDArray:
    original_timestamps = np.linspace(0, duration, num=len(hr))
    target_interval = 1 / target_fs
    target_times = np.arange(0, duration, target_interval)
    interp_func = interp1d(original_timestamps, hr, kind='linear')
    interpolated_hr = interp_func(target_times)
    return interpolated_hr


def compute_hr_from_xxi(positions, xxis, sample_rate, n_samples, window_sec=3):
    xxis_ms = xxis / sample_rate * 1000
    filter_xxis_ms = get_nn_intervals(
        xxis_ms, verbose=False, ectopic_beats_removal_method='karlsson')
    filter_xxis_ms = np.array(list(map(lambda x: x / 1000, filter_xxis_ms)))
    ref = {'time': positions / sample_rate, 'iei': filter_xxis_ms}
    totalSec = int(n_samples / sample_rate)

    # Generate evaluation times
    stepSec = 1
    evalSec = np.arange(0, totalSec - window_sec + stepSec, stepSec)
    windows_start = evalSec[:, np.newaxis]
    windows_end = windows_start + window_sec

    # Create a mask for each window
    within_window = (ref['time'] > windows_start) & (ref['time'] <= windows_end)

    # Calculate heart rate for each window
    hr = 60 / np.array([np.mean(ref['iei'][mask]) if mask.any() else np.nan for mask in within_window])

    # Interpolation and NaN handling
    if len(hr) < totalSec:
        fill_length = totalSec - len(hr)
        left_fill_length = int((window_sec - 1) / 2)
        right_fill_length = fill_length - left_fill_length
        left_fill = np.full(left_fill_length, np.nan)
        right_fill = np.full(right_fill_length, np.nan)
        hr = np.concatenate((left_fill, hr, right_fill))

    # Convert to pandas Series and interpolate
    output_hr = pd.Series(hr).interpolate(method='linear').fillna(method='bfill')
    fs = 1 / stepSec
    return output_hr, fs


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   utils 
@Time        :   2024/9/6 19:29
@Author      :   Xuesong Chen
@Description :   
"""

import neurokit2 as nk
import numpy as np


def get_normalized_timefrequency(
        sigal,
        fs,
        seg_duration=2,
        seg_overlap_ratio=0.5,
        min_frequency=0.3,
        max_frequency=35,
        log=False
):
    freq, time, power = nk.signal_timefrequency(
        sigal,
        window=seg_duration,
        overlap=fs * seg_duration * seg_overlap_ratio,
        min_frequency=min_frequency,
        max_frequency=max_frequency,
        sampling_rate=fs,
        show=False)
    if log:
        power = np.log(power + 1)
    norm_power = power / (np.sum(power, axis=0, keepdims=True) + 1e-8)
    return norm_power  # freq * time



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   detect_oxygen_desaturation
@Time        :   2023/4/25 14:41
@Author      :   Xuesong Chen
@Description :
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

SPO2_MIN_VALUE = 110


def detect_oxygen_desaturation(spo2_arr, duration_max=120, spo2_des_min_thre=3, ret_format='df'):
    spo2_max = spo2_arr[0]  # 初始化最大值
    spo2_max_index = 1  # 初始化最大值下标
    spo2_min = SPO2_MIN_VALUE  # 初始化血氧最低值
    des_onset_pred_set = []  # 算法预测氧降起始点总集
    des_duration_pred_set = []  # 算法预测氧降持续时间总集
    des_level_set = []  # 被记录的氧降事件合集
    des_onset_pred_point = 0  # 预测事件起始点
    des_flag = 0  # 氧降事件发生标记
    ma_flag = 0  # motion artifact事件发生标记
    spo2_des_max_thre = 50  # 血氧motion artifact阈值
    duration_min = 5  # 氧降事件至少持续duration_min s才会被记录在内
    prob_end = []

    for i, current_value in enumerate(spo2_arr):
        if np.isnan(current_value):
            # 重置
            spo2_max = current_value
            spo2_max_index = i
            ma_flag = 0
            des_flag = 0
            spo2_min = SPO2_MIN_VALUE
            prob_end = []
            continue
        des_percent = spo2_max - current_value  # 氧降值

        # 检测Motion artifacts
        if ma_flag and (des_percent < spo2_des_max_thre):
            if des_flag and prob_end:
                des_onset_pred_set.append(des_onset_pred_point)
                des_duration_pred_set.append(prob_end[-1] - des_onset_pred_point)
                des_level_point = spo2_max - spo2_min
                des_level_set.append(des_level_point)
            # 重置
            spo2_max = current_value
            spo2_max_index = i
            ma_flag = 0
            des_flag = 0
            spo2_min = SPO2_MIN_VALUE
            prob_end = []
            continue

        # 如果氧降值大于设置的阈值
        if des_percent >= spo2_des_min_thre:
            if des_percent > spo2_des_max_thre:
                ma_flag = 1
            else:
                des_onset_pred_point = spo2_max_index
                des_flag = 1
                if current_value < spo2_min:
                    spo2_min = current_value

        if current_value >= spo2_max and not des_flag:
            spo2_max = current_value
            spo2_max_index = i

        elif des_flag:
            if current_value > spo2_min:
                if current_value > spo2_arr[i - 1]:
                    prob_end.append(i)

                if current_value <= spo2_arr[i - 1] < spo2_arr[i - 2]:
                    spo2_des_duration = prob_end[-1] - spo2_max_index
                    if spo2_des_duration < duration_min:
                        spo2_max = spo2_arr[i - 2]
                        spo2_max_index = i - 2
                        spo2_min = SPO2_MIN_VALUE
                        des_flag = 0
                        prob_end = []
                        continue
                    else:
                        if duration_min <= spo2_des_duration <= duration_max:
                            des_onset_pred_set.append(des_onset_pred_point)
                            des_duration_pred_set.append(spo2_des_duration)
                            des_level_point = spo2_max - spo2_min
                            des_level_set.append(des_level_point)
                        else:
                            des_onset_pred_set.append(des_onset_pred_point)
                            des_duration_pred_set.append(prob_end[0] - des_onset_pred_point)
                            des_level_point = spo2_max - spo2_min
                            des_level_set.append(des_level_point)
                            remain_spo2_arr = spo2_arr[prob_end[0]:i + 1]
                            _onset, _duration, _des_level = detect_oxygen_desaturation(remain_spo2_arr, ret_format='tuple')
                            des_onset_pred_set.extend([i + prob_end[0] for i in _onset])
                            des_duration_pred_set.extend(_duration)
                            des_level_set.extend(_des_level)
                        spo2_max = spo2_arr[i - 2]
                        spo2_max_index = i - 2
                        spo2_min = SPO2_MIN_VALUE
                        des_flag = 0
                        prob_end = []

    if ret_format == 'tuple':
        return des_onset_pred_set, des_duration_pred_set, des_level_set
    else:
        return pd.DataFrame({
            'Type': 'OD',
            'Start': des_onset_pred_set,
            'Duration': des_duration_pred_set,
            'OD_level': des_level_set
        })


def calc_dynamic_spo2_burden(arr, window=100, default_value=100.0) -> np.ndarray:
    # 初始化满长度的结果数组，所有值设置为100（因为窗口不足的最大值为100）
    max_values = np.full(arr.shape, default_value)
    arr_replace_nan = np.nan_to_num(arr, nan=default_value)
    # 只有当数组长度大于等于窗口长度时，才进行滑动窗口的最大值计算
    if len(arr_replace_nan) >= window:
        # 创建滑动窗口视图
        shape = arr_replace_nan.shape[:-1] + (
            arr_replace_nan.shape[-1] - window,
            window,
        )
        strides = arr_replace_nan.strides + (arr_replace_nan.strides[-1],)
        windows = np.lib.stride_tricks.as_strided(
            arr_replace_nan, shape=shape, strides=strides
        )
        # 计算每个窗口的最大值
        max_values[window:] = np.max(windows, axis=1)
    # 计算最大值和原数组对应值的差
    result = max_values - arr
    return result.astype(np.float32)


if __name__ == '__main__':
    from wuji.Reader import PhilipsEDFReader
    fp = '/Users/cxs/Downloads/00001221-113080/00001221-113080_1087800.edf'
    reader = PhilipsEDFReader(fp)
    gt_spo2 = reader.get_signal('SpO2')
    raw_spo2 = reader.get_signal('SpO2')
    fs = reader.get_sample_frequency('SpO2')
    from wuji.tools.plot_labeled_signal import plot_signal_with_events
    from  wuji.Preprocessor.signal import filter_spo2
    # from wuji.algo.Oxygen.desaturation_wrapper import detect_desaturation
    gt_spo2 = filter_spo2(gt_spo2, fs=1)
    gt_spo2 = np.nan_to_num(gt_spo2, nan=-100)
    # gt_spo2 = calc_dynamic_spo2_burden(gt_spo2, 100)
    # gt_spo2 = np.nan_to_num(90 - gt_spo2, nan=-100)
    plt.plot(gt_spo2)
    plt.show()

    import time
    s_t = time.time()
    i = 0
    while i < 1:
        res = detect_oxygen_desaturation(gt_spo2)
        i += 1
    print(time.time() - s_t)
    print(res)
    plt.plot(raw_spo2)
    plot_signal_with_events(gt_spo2, 1, res)
    # plt.plot(gt_spo2)
    # plt.show()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utilities
==================

Functions to handle event lists (list of event items), event rolls (event activity indicator matrix used in evaluation),
and scene lists.

Event list operations
---------------------

.. autosummary::
    :toctree: generated/

    event_list.unique_event_labels
    event_list.unique_files
    event_list.filter_event_list
    event_list.max_event_offset

Event roll operations
---------------------

.. autosummary::
    :toctree: generated/

    event_roll.event_list_to_event_roll
    event_roll.pad_event_roll
    event_roll.match_event_roll_lengths

Scene list operations
---------------------

.. autosummary::
    :toctree: generated/

    scene_list.unique_scene_labels

Event matching operations
-------------------------

.. autosummary::
    :toctree: generated/

    event_matching.bipartite_match

"""

from .event_list import *
from .event_roll import *
from .scene_list import *
from .event_matching import *

__all__ = [_ for _ in dir() if not _.startswith('_')]


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   bcg2timefreq 
@Time        :   2024/9/6 19:26
@Author      :   Xuesong Chen
@Description :   
"""
import numpy as np

from wuji.biosignal_transformer.utils import get_normalized_timefrequency


def get_bcg_feature(
        sigal,
        fs,
        seg_duration=0.5,
        seg_overlap_ratio=0.25,
        min_frequency=0.1,
        max_frequency=25,
):
    return get_normalized_timefrequency(
        sigal,
        fs,
        seg_duration=seg_duration,
        seg_overlap_ratio=seg_overlap_ratio,
        min_frequency=min_frequency,
        max_frequency=max_frequency,
        log=True
    )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   NSRR 
@Time        :   2023/8/17 16:47
@Author      :   Xuesong Chen
@Description :   
"""
import os

import xmltodict
import pandas as pd

from wuji.Reader.Annotation.Base import Base, AHI, AHEventFilter
from wuji.Reader.utils import get_equal_duration_and_labeled_chunks
from datetime import datetime


class NSRRAnnotationReader(Base):
    def __init__(self, file_path):
        super().__init__(file_path)

    def _parse_file(self, file_path):
        with open(file_path, encoding='utf-8') as f:
            info_dict = xmltodict.parse(f.read())
            self.scored_events = pd.DataFrame(info_dict['PSGAnnotation']['ScoredEvents']['ScoredEvent'])
            date_string = self.scored_events.loc[0, 'ClockTime']
            fake_datetime = datetime.strptime('2000-01-01 ' + date_string.split(' ')[-1], '%Y-%m-%d %H.%M.%S')
            self.recording_start_time = fake_datetime
            self.duration = float(self.scored_events.loc[0, 'Duration'])
            self.scored_events[['Start', 'Duration']] = self.scored_events[['Start', 'Duration']].astype(float)
            self.scored_events = self.scored_events.iloc[1:]

    def get_standard_sleep_stages(self):
        stages = self.scored_events[self.scored_events['EventType'] == 'Stages|Stages'].copy()
        stages.loc[:, 'stage_num'] = stages['EventConcept'].str.split('|', expand=True)[1].astype(int)
        map_dic = {0: 'Wake', 1: 'N1', 2: 'N2', 3: 'N3', 4: 'N3', 5: 'REM'}
        stages.loc[:, 'Type'] = stages['stage_num'].map(map_dic)
        stages = stages[['Type', 'Start', 'Duration']]
        standard_stages = get_equal_duration_and_labeled_chunks(stages)
        self.sleep_stages = standard_stages
        return standard_stages

    def get_standard_AH_events(self, type='AHI3', od_eps=45, aro_eps=6):
        if type == 'AHI4':
            self.get_OD_events(OD_level=4)
            self.arousal_events = None
        elif type == 'AHI3':
            self.get_OD_events(OD_level=3)
            self.get_arousal_events()
        if self.respiratory_events is None:
            self.get_respiratory_events()
        filter = AHEventFilter(self.respiratory_events, self.OD_events, self.arousal_events, self.sleep_stages)
        self.OD_events = self.OD_events[['Type', 'Start', 'Duration', 'OD_level']]
        self.arousal_events = self.arousal_events[['Type', 'Start', 'Duration']]
        return filter.get_filtered_AH_events(type=type, od_eps=od_eps, aro_eps=aro_eps)

    def get_respiratory_events(self):
        res = self.scored_events[(self.scored_events['EventConcept'].str.contains('pnea'))
                                 | (self.scored_events['EventConcept'] == 'Unsure|Unsure')].copy()
        EventConcept_map = {
            'Hypopnea|Hypopnea': 'Hypopnea',
            'Obstructive apnea|Obstructive Apnea': 'Apnea',
            'Unsure|Unsure': 'Hyponpea',
            'Central apnea|Central Apnea': 'Apnea',
            'Mixed apnea|Mixed Apnea': 'Apnea',
        }
        res['Type'] = res['EventConcept'].map(EventConcept_map)
        assert res['Type'].isna().sum() == 0, f"存在未map的呼吸事件:{res['EventConcept'].unique()}"
        self.respiratory_events = res[['Type', 'Start', 'Duration']][(res['Duration'] >= 10) & (res['Duration'] <= 120)]
        self.respiratory_events.reset_index(drop=True, inplace=True)
        return self.respiratory_events

    def get_OD_events(self, OD_level=4):
        res = self.scored_events[(self.scored_events['EventConcept'] != 'Wake|0') &
                                 (self.scored_events['EventConcept'] == 'SpO2 desaturation|SpO2 desaturation')].copy()
        if res.empty:
            self.OD_events = pd.DataFrame(columns=['Type', 'Start', 'Duration', 'OD_level'])
            return self.OD_events
        EventConcept_map = {
            'SpO2 desaturation|SpO2 desaturation': 'OD',
        }
        res['Type'] = res['EventConcept'].map(EventConcept_map)
        res['OD_level'] = res['SpO2Baseline'].astype(float) - res['SpO2Nadir'].astype(float)
        self.OD_events = res[['Type', 'Start', 'Duration', 'OD_level']][(res['OD_level'] >= OD_level)]
        self.OD_events.reset_index(drop=True, inplace=True)
        return self.OD_events

    def get_arousal_events(self):
        res = self.scored_events[(self.scored_events['EventConcept'] != 'Wake|0') &
                                 (self.scored_events['EventConcept'] == 'Arousal|Arousal ()')].copy()
        EventConcept_map = {
            'Arousal|Arousal ()': 'Arousal',
        }
        res['Type'] = res['EventConcept'].map(EventConcept_map)
        self.arousal_events = res[['Type', 'Start', 'Duration']]
        self.arousal_events.reset_index(drop=True, inplace=True)
        return self.arousal_events


if __name__ == '__main__':
    import warnings
    from wuji.Reader.EDF.NSRR import NSRREDFReader

    warnings.filterwarnings('ignore')
    for file in sorted(os.listdir('/Users/cxs/project/OSAPillow/data/SHHS/annotations')):
        if file.endswith('.xml'):
            fp = os.path.join('/Users/cxs/project/OSAPillow/data/SHHS/annotations', file)
            reader = NSRRAnnotationReader(fp)
            stages = reader.get_standard_sleep_stages()
            AH_events = reader.get_standard_AH_events(type='AHI3')
            ahi_calculator = AHI(AH_events, stages)
            ahi = ahi_calculator.get_AHI(type='Total')
            print(f'{file} AHI: {round(ahi, 2)}')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
@File        :   setup 
@Time        :   2023/9/14 15:51
@Author      :   Xuesong Chen
@Description :   
"""
from setuptools import setup, find_packages

setup(
    name='wuji',
    version='0.1.89',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    install_requires=[
        # list of dependencies, e.g., 'numpy>=1.18.5'
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Event list handling
"""

import dcase_util

__all__ = ['filter_event_list',
           'unique_files',
           'unique_event_labels',
           'max_event_offset']


def filter_event_list(event_list, scene_label=None, event_label=None, filename=None):
    """Filter event list based on given fields

    Parameters
    ----------
    event_list : list, shape=(n,)
        A list containing event dicts

    scene_label : str
        Scene label

    event_label : str
        Event label

    filename : str
        Filename

    Returns
    -------
    event_list: list, shape=(n,)
        A list containing event dicts

    """

    return dcase_util.containers.MetaDataContainer(event_list).filter(
        filename=filename,
        scene_label=scene_label,
        event_label=event_label
    )


def unique_files(event_list):
    """Find the unique files

    Parameters
    ----------
    event_list : list or dcase_util.containers.MetaDataContainer
        A list containing event dicts

    Returns
    -------
    list
        Unique filenames in alphabetical order

    """

    if isinstance(event_list, dcase_util.containers.MetaDataContainer):
        return event_list.unique_files

    else:
        files = {}
        for event in event_list:
            if 'file' in event:
                files[event['file']] = event['file']

            elif 'filename' in event:
                files[event['filename']] = event['filename']

        files = list(files.keys())
        files.sort()
        return files


def unique_event_labels(event_list):
    """Find the unique event labels

    Parameters
    ----------
    event_list : list or dcase_util.containers.MetaDataContainer
        A list containing event dicts

    Returns
    -------
    list
        Unique labels in alphabetical order

    """

    if isinstance(event_list, dcase_util.containers.MetaDataContainer):
        return event_list.unique_event_labels

    else:
        labels = []
        for event in event_list:
            if 'event_label' in event and event['event_label'] not in labels:
                labels.append(event['event_label'])

        labels.sort()
        return labels


def max_event_offset(event_list):
    """Find the offset (end-time) of last event

    Parameters
    ----------
    event_list : list or dcase_util.containers.MetaDataContainer
        A list containing event dicts

    Returns
    -------
    float > 0
        maximum offset

    """

    if isinstance(event_list, dcase_util.containers.MetaDataContainer):
        return event_list.max_offset

    else:
        max_offset = 0
        for event in event_list:
            if 'event_offset' in event:
                if event['event_offset'] > max_offset:
                    max_offset = event['event_offset']

            elif 'offset' in event:
                if event['offset'] > max_offset:
                    max_offset = event['offset']

        return max_offset
